from pepnet import Predictor, SequenceInput, Output
from nose.tools import eq_
import numpy as np

def test_predictor_json_identity():
    predictor = Predictor(
        inputs=[SequenceInput(length=2, variable_length=False, encoding="onehot")],
        outputs=[Output(dim=1, activation="sigmoid")])
    eq_(predictor, Predictor.from_json(predictor.to_json()))

def test_predictor_weights_all_ones():
    predictor = Predictor(
        inputs=[SequenceInput(length=2, variable_length=False, encoding="onehot")],
        outputs=[Output(dim=1, activation="sigmoid")])
    weights = predictor.get_weights()
    for w in weights:
        w.fill(1)
    predictor.set_weights(weights)
    predictor2 = Predictor.from_json(predictor.to_json())
    for w in predictor2.get_weights():
        assert (w == np.ones_like(w)).all(), "Expected %s to be all 1s" % (w,)



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .nn_helpers import dense_layers, make_numeric_input
from .numeric import Numeric

class NumericInput(Numeric):
    """
    Input which expects fixed length vector, takes same arguments as
    NumericOutput (defined in base class Numeric).
    """
    def build(self):
        input_object = make_numeric_input(
            name=self.name, dim=self.dim, dtype=self.dtype)
        hidden = dense_layers(
            value=input_object,
            layer_sizes=self.dense_layer_sizes,
            activation=self.dense_activation,
            dropout=self.dense_dropout,
            batch_normalization=self.dense_batch_normalization)
        return input_object, hidden



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .keras_layers.masked_maxpooling1d import MaskedMaxPooling1D as MaxPooling1D
from .keras_layers.masked_conv1d import MaskedConv1D as Conv1D
from .keras_layers.masked_global_average_pooling import (
    MaskedGlobalAveragePooling1D as GlobalAveragePooling1D)
from .keras_layers.masked_global_max_pooling import (
    MaskedGlobalMaxPooling1D as GlobalMaxPooling1D)
from .keras_layers.drop_mask import DropMask

from keras.layers import (
    Input,
    Embedding,
    Dropout,
    SpatialDropout1D,
    Dense,
    Activation,
    BatchNormalization,
    Concatenate,
    Reshape,
    Multiply,
    Add,
    Flatten,
    Lambda,
    LSTM,
    GRU,
    Bidirectional
)

import keras.backend as K
import keras.initializers

def make_onehot_sequence_input(name, length, n_symbols):
    return Input(
            shape=(length, n_symbols),
            name=name,
            dtype="float32")

def make_index_sequence_input(name, length):
    return Input(
        shape=(length,),
        dtype="int32",
        name=name)

def make_sequence_input(name, length, n_symbols, encoding):
    assert encoding in {"onehot", "index"}
    if encoding == "index":
        return make_index_sequence_input(name, length)
    else:
        return make_onehot_sequence_input(name, length, n_symbols)

def make_numeric_input(name, dim, dtype):
    return Input(name=name, shape=(dim,), dtype=dtype)


def merge(values, merge_mode):
    assert merge_mode in {"concat", "add", "multiply"}
    if len(values) == 1:
        return values[0]
    elif merge_mode == "concat":
        return Concatenate()(values)
    elif merge_mode == "add":
        return Add()(values)
    elif merge_mode == "multiply":
        return Multiply()(values)

def flatten(value, drop_mask=False):
    if drop_mask:
        value = DropMask()(value)
    return Flatten()(value)

def regularize(value, batch_normalization=False, dropout=0.0):
    if batch_normalization:
        value = BatchNormalization()(value)
    if dropout:
        value = Dropout(dropout)(value)
    return value

def embedding(
        value,
        n_symbols,
        output_dim,
        dropout=0,
        initial_weights=None,
        mask_zero=False):
    if initial_weights:
        n_rows, n_cols = initial_weights.shape
        if n_rows != n_symbols or n_cols != output_dim:
            raise ValueError(
                "Wrong shape for embedding: expected (%d, %d) but got "
                "(%d, %d)" % (
                    n_symbols, output_dim,
                    n_rows, n_cols))
        embedding_layer = Embedding(
            input_dim=n_symbols,
            output_dim=output_dim,
            mask_zero=mask_zero,
            weights=[initial_weights])
    else:
        embedding_layer = Embedding(
            input_dim=n_symbols,
            output_dim=output_dim,
            mask_zero=mask_zero)

    value = embedding_layer(value)

    if dropout:
        value = SpatialDropout1D(dropout)(value)

    return value

def parametric_conv(
        value,
        weight_source,
        filter_size,
        output_dim,
        padding):
    assert False, "Halp, how to make separate convs per sample"
    n_timesteps, input_dim = K.int_shape(value)[-2:]
    kernel_shape = (filter_size, input_dim, output_dim)
    kernel = dense(
        weight_source,
        dim=kernel_shape[0] * kernel_shape[1] * kernel_shape[1],
        activation="linear")
    kernel = Reshape(kernel_shape)(kernel)
    return K.conv1d(
        value, kernel, padding=padding)

def conv(
        value,
        filter_size,
        output_dim,
        dropout=0.0,
        activation="linear",
        padding="valid",
        weight_source=None):
    """
    Perform a single scale of convolution and optionally add spatial dropout.
    """
    if weight_source is None:
        conv_layer = Conv1D(
            filters=output_dim,
            kernel_size=filter_size,
            padding=padding,
            activation=activation)
        convolved = conv_layer(value)
    else:
        convolved = parametric_conv(
            value,
            weight_source,
            filter_size=filter_size,
            output_dim=output_dim,
            padding=padding)
    if dropout > 0:
        # random drop some of the convolutional filters
        convolved = SpatialDropout1D(dropout)(convolved)
    return convolved

def aligned_convolutions(
        value,
        filter_sizes,
        output_dim,
        activation="linear",
        dropout=0.0,
        batch_normalization=False,
        weight_source=None):
    """
    Perform convolutions at multiple scales and concatenate their outputs.

    Parameters
    ----------
    filter_sizes : int or list of int
        Widths of convolutional filters

    output_dim : int
        Number of filters per width

    dropout : float
        Dropout after convolutional

    weight_source : tensor, optional
        Compute weights as a function of existing tensor

    """
    if isinstance(filter_sizes, int):
        filter_sizes = [filter_sizes]
    if isinstance(output_dim, dict):
        given_sizes = set(output_dim.keys())
        if given_sizes != set(filter_sizes):
            raise ValueError("Expected filter sizes %s but got %s" % (
                set(filter_sizes), given_sizes))
    else:
        assert isinstance(output_dim, int)
        output_dim = {size: output_dim for size in filter_sizes}

    convolved_list = []
    for size in filter_sizes:
        convolved_list.append(conv(
            value=value,
            filter_size=size,
            output_dim=output_dim[size],
            padding="same",
            activation=activation,
            weight_source=weight_source,
            dropout=0))
    convolved = merge(convolved_list, "concat")
    if batch_normalization:
        convolved = BatchNormalization()(convolved)
    if dropout > 0:
        # random drop some of the convolutional filters
        convolved = SpatialDropout1D(dropout)(convolved)
    return convolved

def local_max_pooling(value, size=3, stride=2):
    return MaxPooling1D(pool_size=size, strides=stride)(value)

def global_max_pooling(value, batch_normalization=False, dropout=0):
    return regularize(
        value=GlobalMaxPooling1D()(value),
        batch_normalization=batch_normalization,
        dropout=dropout)

def global_mean_pooling(value, batch_normalization=False, dropout=0):
    return regularize(
        value=GlobalAveragePooling1D()(value),
        batch_normalization=batch_normalization,
        dropout=dropout)

def global_max_and_mean_pooling(
        value, batch_normalization=False, dropout=0):
    return merge([
        global_max_pooling(
            value=value,
            batch_normalization=batch_normalization,
            dropout=dropout),
        global_mean_pooling(
            value=value,
            batch_normalization=batch_normalization,
            dropout=dropout)], "concat")

def dense(value, dim, activation, init="glorot_uniform", name=None):
    if name:
        # hidden layer fully connected layer
        value = Dense(
            units=dim, kernel_initializer=init, name="%s_dense" % name)(value)
        value = Activation(activation, name=name)(value)
    else:
        value = Dense(units=dim, kernel_initializer=init)(value)
        value = Activation(activation)(value)
    return value

def dense_layers(
        value,
        layer_sizes,
        activation="relu",
        init="glorot_uniform",
        batch_normalization=False,
        dropout=0.0):
    for i, dim in enumerate(layer_sizes):
        value = regularize(
            value=dense(value, dim=dim, init=init, activation=activation),
            batch_normalization=batch_normalization,
            dropout=dropout)
    return value


def highway_layer(value, activation="tanh", gate_bias=-3):
    dims = K.int_shape(value)
    if len(dims) != 2:
        raise ValueError(
            "Expected 2d value (batch_size, dims) but got shape %s" % (
                dims,))
    dim = dims[-1]
    gate_bias_initializer = keras.initializers.Constant(gate_bias)
    gate = Dense(units=dim, bias_initializer=gate_bias_initializer)(value)
    gate = Activation("sigmoid")(gate)
    negated_gate = Lambda(
        lambda x: 1.0 - x,
        output_shape=(dim,))(gate)
    transformed = Dense(units=dim)(value)
    transformed = Activation(activation)(value)
    transformed_gated = Multiply()([gate, transformed])
    pass_through_gated = Multiply()([negated_gate, value])
    value = Add()([transformed_gated, pass_through_gated])
    return value

def highway_layers(
        value,
        n_layers,
        activation="tanh",
        batch_normalization=False,
        dropout=0,
        gate_bias=-3):
    """
    Construct "highway" layers which default to the identity function
    but can learn to transform their input. The batch normalization
    and dropout parameters only affect the output of the last layer.
    """
    for i in range(n_layers):
        value = highway_layer(
            value, activation=activation, gate_bias=gate_bias)
    return regularize(
        value=value,
        batch_normalization=batch_normalization,
        dropout=dropout)

def recurrent_layers(
        value,
        layer_sizes,
        bidirectional=True,
        dropout=0.0,
        rnn_type="lstm"):
    """
    Make one or more RNN layers
    """
    if rnn_type == "lstm":
        rnn_class = LSTM
    elif rnn_type == "gru":
        rnn_class = GRU
    else:
        raise ValueError("Unknown RNN type: %s" % (rnn_type,))

    for i, layer_size in enumerate(layer_sizes):
        last_layer = (i == len(layer_sizes) - 1)
        rnn_layer = rnn_class(
            layer_size,
            return_sequences=not last_layer,
            dropout=dropout)
        if bidirectional:
            rnn_layer = Bidirectional(rnn_layer, merge_mode="concat")
        value = rnn_layer(value)
    return value

def tensor_shape(x):
    return K.int_shape(x)



import numpy

from ..amino_acids import amino_acids_dict

AMINO_ACIDS = list(amino_acids_dict)


def random_peptides(num, length=9):
    """
    Generate uniformly random peptides (kmers).

    Parameters
    ----------
    num : int
        Number of peptides to return

    length : int
        Length of each peptide

    Returns
    ----------
    list of string

    """
    if num == 0:
        return []
    return [
        ''.join(peptide_sequence)
        for peptide_sequence in
        numpy.random.choice(
            AMINO_ACIDS, size=(int(num), int(length)))
    ]


# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from serializable import Serializable

from .nn_helpers import (
    aligned_convolutions,
    embedding,
    make_sequence_input,
    local_max_pooling,
    global_max_and_mean_pooling,
    flatten,
    recurrent_layers,
    highway_layers,
    dense_layers
)
from .encoder import Encoder

class SequenceInput(Serializable):
    def __init__(
            self,
            length,
            name=None,
            variable_length=False,
            # embedding of symbol indices into vectors
            encoding="index",
            add_start_tokens=False,
            add_stop_tokens=False,
            # embedding
            embedding_dim=32,
            embedding_dropout=0,
            # convolutional layers
            conv_filter_sizes=[],
            repeat_conv_layers=1,
            conv_dropout=0,
            conv_batch_normalization=False,
            conv_activation="linear",
            conv_weight_source=None,
            pool_size=3,
            pool_stride=2,
            # RNN
            rnn_layer_sizes=[],
            rnn_type="lstm",
            rnn_bidirectional=True,
            # global pooling of conv or RNN outputs
            global_pooling=False,
            global_pooling_batch_normalization=False,
            global_pooling_dropout=0,
            # transform global pooling representation with
            # dense layers
            dense_layer_sizes=[],
            dense_activation="relu",
            dense_dropout=0.25,
            dense_batch_normalization=False,
            # highway layers
            n_highway_layers=0,
            highway_activation="tanh",
            highway_batch_normalization=False,
            highway_dropout=0,):
        """
        Parameters
        ----------
        length : int
            Maximum length of sequence

        name : str
            Name of input sequence

        variable_length : bool
            Do we expect padding '-' characters in the input?

        encoding : {"index", "onehot"}
            How are symbols represented: via integer indices or boolean
            vectors?

        add_start_tokens : bool
            Add "^" token to start of each sequence

        add_stop_tokens : bool
            Add "$" token to end of each sequence

        embedding_dim : int
            How many dimensions in the symbol embedding
            (only used for index encoding)

        embedding_dropout : float
            What fraction of symbol representations are randomly set to 0
            during training?

        conv_filter_sizes : list of dict
            List whose elements describe to convolutional layers. Each
            element of the list is a dictionary whose keys are filter widths
            and whose values are the number of filters associated with that
            width.

        repeat_conv_layers : int
            Number of times to repeat convolutional layers

        conv_dropout : float
            Fraction of convolutional activations to randomly set to 0 during
            training

        conv_batch_normalization : bool
            Apply batch normalization between convolutional layers

        conv_weight_source : tensor, optional
            Determine weights of the convolution as a function of this
            tensor (rather than learning parameters separately)

        pool_size : int
            If using more than one convolutional layer, how many timesteps
            from a previous layer get combined via maxpooling before the next
            layer.

        pool_stride : int
            If using more than one convolutional layer, stride of the max
            pooling.

        rnn_layer_sizes : list of int
            Dimensionality of each RNN layer, defaults no RNN layers

        rnn_type : {"lstm", "gru"}
            Whether to use GRU or LSTM models for each RNN layer

        rnn_bidirectional : bool
            Use bidirectional RNNs

        global_pooling : bool
            Pool (mean & max) activations across sequence length

        global_pooling_batch_normalization : bool
            Apply BatchNormalization after global pooling

        global_pooling_dropout : float
            Fraction of entries to randomly drop during training after
            global pooling

        dense_layer_sizes : list of int
            Dimensionality of dense transformations after convolutional
            and recurrent layers

        dense_activation: str
            Activation function to use after each dense layer

        dense_dropout : float
            Fraction of dense output values to drop during training

        dense_batch_normalization : bool
            Apply batch normalization between hidden layers

        n_highway_layers : int
            Number of highway layers to use after dense layers

        highway_batch_normalization : bool
            Apply BatchNormalization after final highway layer

        highway_dropout : float
            Apply dropout after final highway layer

        highway_activation : str
            Activation function of each layer in the highway network
        """
        self.name = name
        self.length = length
        if encoding not in {"index", "onehot"}:
            raise ValueError("Invalid encoding: %s" % encoding)
        self.encoding = encoding
        self.add_start_tokens = add_start_tokens
        self.add_stop_tokens = add_stop_tokens
        self.variable_length = variable_length
        self.encoder = Encoder(
            variable_length_sequences=self.variable_length,
            add_start_tokens=self.add_start_tokens,
            add_stop_tokens=self.add_stop_tokens)
        self.n_symbols = len(self.encoder.tokens)

        self.embedding_dim = embedding_dim
        self.embedding_dropout = embedding_dropout

        self.conv_filter_sizes = conv_filter_sizes
        self.repeat_conv_layers = repeat_conv_layers
        self.conv_dropout = conv_dropout
        self.conv_batch_normalization = conv_batch_normalization
        self.conv_activation = conv_activation
        self.conv_weight_source = conv_weight_source
        self.pool_size = pool_size
        self.pool_stride = pool_stride

        self.rnn_layer_sizes = rnn_layer_sizes
        self.rnn_type = rnn_type
        self.rnn_bidirectional = rnn_bidirectional

        self.global_pooling = global_pooling
        self.global_pooling_batch_normalization = global_pooling_batch_normalization
        self.global_pooling_dropout = global_pooling_dropout

        # Dense layers after all temporal processing
        self.dense_layer_sizes = dense_layer_sizes
        self.dense_activation = dense_activation
        self.dense_dropout = dense_dropout
        self.dense_batch_normalization = dense_batch_normalization

        # Highway network after dense layers
        self.n_highway_layers = n_highway_layers
        self.highway_batch_normalization = highway_batch_normalization
        self.highway_dropout = highway_dropout
        self.highway_activation = highway_activation

    def _build_input(self):
        return make_sequence_input(
            encoding=self.encoding,
            name=self.name,
            length=self.length + self.add_start_tokens + self.add_stop_tokens,
            n_symbols=self.n_symbols)

    def _build_embedding(self, input_object):
        if self.encoding == "index":
            if self.embedding_dim <= 0:
                raise ValueError(
                    "Invalid embedding dim: %d" % self.embedding_dim)
            return embedding(
                input_object,
                n_symbols=self.n_symbols,
                output_dim=self.embedding_dim,
                mask_zero=self.variable_length,
                dropout=self.embedding_dropout)
        else:
            return input_object

    def _build_conv(self, value):
        if self.conv_filter_sizes:
            if isinstance(self.conv_filter_sizes, dict):
                # if only one dictionary is given, then treat it as a single
                # layer
                conv_filter_sizes = [self.conv_filter_sizes]
            else:
                conv_filter_sizes = self.conv_filter_sizes
            conv_weight_source = self.conv_weight_source
            if conv_weight_source is not None and isinstance(
                    conv_weight_source, SequenceInput):
                conv_weight_source = conv_weight_source.build()[1]

            conv_layer_index = 0
            n_conv_layers = self.repeat_conv_layers * len(conv_filter_sizes)
            for _ in range(self.repeat_conv_layers):
                for conv_layer_dict in conv_filter_sizes:
                    if not isinstance(conv_layer_dict, dict):
                        raise ValueError((
                            "Each element of conv_filter_sizes must be a "
                            "{width: num_filters} dictionary, "
                            "got %s : %s instead." % (
                                conv_layer_dict, type(conv_layer_dict))))
                    elif len(conv_layer_dict) == 0:
                        continue
                    value = aligned_convolutions(
                        value,
                        filter_sizes=list(conv_layer_dict.keys()),
                        output_dim=conv_layer_dict,
                        dropout=self.conv_dropout,
                        batch_normalization=self.conv_batch_normalization,
                        activation=self.conv_activation,
                        weight_source=conv_weight_source)
                    conv_layer_index += 1
                    if conv_layer_index < n_conv_layers:
                        # add max pooling for all layers before the last
                        value = local_max_pooling(
                            value,
                            size=self.pool_size,
                            stride=self.pool_stride)
        return value

    def _build_rnn(self, value):
        if isinstance(self.rnn_layer_sizes, int):
            rnn_layer_sizes = [self.rnn_layer_sizes]
        else:
            rnn_layer_sizes = self.rnn_layer_sizes

        if len(rnn_layer_sizes) > 0:
            value = recurrent_layers(
                value=value,
                layer_sizes=rnn_layer_sizes,
                rnn_type=self.rnn_type,
                bidirectional=self.rnn_bidirectional)
        return value

    def _build_global_pooling(self, value):
        if self.global_pooling:
            value = global_max_and_mean_pooling(
                value,
                batch_normalization=self.global_pooling_batch_normalization,
                dropout=self.global_pooling_dropout)
        return value

    def _build_dense(self, value):
        if value.ndim > 2:
            value = flatten(value, drop_mask=self.variable_length)

        value = dense_layers(
            value,
            layer_sizes=self.dense_layer_sizes,
            activation=self.dense_activation,
            dropout=self.dense_dropout,
            batch_normalization=self.dense_batch_normalization)
        return value

    def _build_highway(self, value):
        if value.ndim > 2:
            value = flatten(value, drop_mask=self.variable_length)
        if self.n_highway_layers:
            value = highway_layers(
                value,
                activation=self.highway_activation,
                n_layers=self.n_highway_layers,
                dropout=self.highway_dropout,
                batch_normalization=self.highway_batch_normalization)
        return value

    def build(self):
        input_object = self._build_input()
        value = self._build_embedding(input_object)
        value = self._build_conv(value)
        value = self._build_rnn(value)
        value = self._build_global_pooling(value)
        value = self._build_dense(value)
        value = self._build_highway(value)
        return input_object, value

    def encode(self, peptides):
        if self.encoding == "index":
            return self.encoder.encode_index_array(
                peptides,
                max_peptide_length=self.length)
        else:
            return self.encoder.encode_onehot(
                peptides, max_peptide_length=self.length)



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from collections import OrderedDict

from .amino_acids import amino_acids_dict


class Encoder(object):
    """
    Container for mapping between amino acid letter codes and their full names
    and providing index/hotshot encodings of amino acid sequences.

    The reason we need a class to contain these mappins is that we might want to
    operate on just the common 20 amino acids or a richer set with modifications.
    """
    def __init__(
            self,
            amino_acid_alphabet=amino_acids_dict,
            variable_length_sequences=True,
            add_start_tokens=False,
            add_stop_tokens=False):
        """
        Parameters
        ----------
        tokens_to_names_dict : dict
            Dictionary mapping each amino acid to its name

        variable_length_sequences : bool
            Do we expect to encode peptides of varying lengths? If so, include
            the gap token "-" in the encoder's alphabet.

        add_start_tokens : bool
            Prefix each peptide string with "^"

        add_stop_tokens : bool
            End each peptide string with "$"
        """
        self._tokens_to_names = OrderedDict()
        self._index_dict = {}

        self.amino_acid_alphabet = amino_acid_alphabet
        self.variable_length_sequences = variable_length_sequences
        self.add_start_tokens = add_start_tokens
        self.add_stop_tokens = add_stop_tokens

        if self.variable_length_sequences:
            self._add_token("-", "Gap")

        if self.add_start_tokens:
            self._add_token("^", "Start")

        if self.add_stop_tokens:
            self._add_token("$", "Stop")

        for (k, v) in amino_acid_alphabet.items():
            self._add_token(k, v)

    def _add_token(self, token, name):
        assert len(token) == 1, "Invalid token '%s'" % (token,)
        assert token not in self._index_dict
        assert token not in self._tokens_to_names
        self._index_dict[token] = len(self._index_dict)
        self._tokens_to_names[token] = name

    def prepare_sequences(self, peptides, padded_peptide_length=None):
        """
        Add start/stop tokens to each peptide (if required) and
        if padded_peptide_length is provided then pad each peptide to
        be the same length using the gap token '-'.
        """
        if self.add_start_tokens:
            peptides = ["^" + p for p in peptides]
            if padded_peptide_length:
                padded_peptide_length += 1

        if self.add_stop_tokens:
            peptides = [p + "$" for p in peptides]
            if padded_peptide_length:
                padded_peptide_length += 1

        if padded_peptide_length:
            peptides = [
                p + "-" * (padded_peptide_length - len(p))
                for p in peptides
            ]
        return peptides

    @property
    def tokens(self):
        """
        Return letters in sorted order, special characters should get indices
        lower than actual amino acids in this order:
            1) "-"
            2) "^"
            3) "$"
        Currently we're enforcing this order by having the _tokens_to_names
        dictionary be an OrderedDict and adding special tokens before amino
        acids in the __init__ method.
        """
        return list(self._tokens_to_names.keys())

    @property
    def amino_acid_names(self):
        return [self._tokens_to_names[k] for k in self.tokens]

    @property
    def index_dict(self):
        return self._index_dict

    def __getitem__(self, k):
        return self._tokens_to_names[k]

    def __setitem__(self, k, v):
        self._add_token(k, v)

    def __len__(self):
        return len(self.tokens)

    def _validate_peptide_lengths(
            self,
            peptides,
            max_peptide_length=None):
        if max_peptide_length is None:
            max_peptide_length = max(len(p) for p in peptides)

        if self.variable_length_sequences:
            max_observed_length = max(len(p) for p in peptides)
            if max_observed_length > max_peptide_length:
                example = [p for p in peptides if len(p) == max_observed_length][0]
                raise ValueError(
                    "Can't have peptide(s) of length %d and max_peptide_length = %d (example '%s')" % (
                        max_observed_length,
                        max_peptide_length,
                        example))
        elif any(len(p) != max_peptide_length for p in peptides):
            example = [p for p in peptides if len(p) != max_peptide_length][0]
            raise ValueError("Expected all peptides to have length %d, '%s' has length %d" % (
                max_peptide_length,
                example,
                len(example)))
        return max_peptide_length

    def _validate_and_prepare_peptides(self, peptides, max_peptide_length=None):
        max_peptide_length = self._validate_peptide_lengths(
            peptides, max_peptide_length)
        peptides = self.prepare_sequences(peptides)
        # did we add start tokens to each sequence?
        max_peptide_length += self.add_start_tokens
        # did we add stop tokens to each sequence?
        max_peptide_length += self.add_stop_tokens
        return peptides, max_peptide_length

    def encode_index_lists(self, peptides):
        # don't try to do length validation since we're allowed to have
        # multiple peptide lengths
        peptides = self.prepare_sequences(peptides)
        index_dict = self.index_dict
        return [
            [index_dict[amino_acid] for amino_acid in peptide]
            for peptide in peptides
        ]

    def encode_index_array(
            self,
            peptides,
            max_peptide_length=None):
        """
        Encode a set of equal length peptides as a matrix of their
        amino acid indices.
        """
        peptides, max_peptide_length = self._validate_and_prepare_peptides(
            peptides, max_peptide_length)
        n_peptides = len(peptides)
        X_index = np.zeros((n_peptides, max_peptide_length), dtype=int)
        index_dict = self.index_dict
        for i, peptide in enumerate(peptides):
            for j, amino_acid in enumerate(peptide):
                # we're expecting the token '-' to have index 0 so it's
                # OK to only loop until the end of the given sequence
                X_index[i, j] = index_dict[amino_acid]
        return X_index

    def encode_onehot(
            self,
            peptides,
            max_peptide_length=None):
        """
        Encode a set of equal length peptides as a binary matrix,
        where each letter is transformed into a length 20 vector with a single
        element that is 1 (and the others are 0).
        """
        peptides, max_peptide_length = self._validate_and_prepare_peptides(
            peptides, max_peptide_length)
        index_dict = self.index_dict
        n_symbols = len(index_dict)
        X = np.zeros((len(peptides), max_peptide_length, n_symbols), dtype=bool)
        for i, peptide in enumerate(peptides):
            for j, amino_acid in enumerate(peptide):
                X[i, j, index_dict[amino_acid]] = 1
        return X

    def encode_FOFE(self, peptides, alpha=0.7, bidirectional=False):
        """
        Implementation of FOFE encoding from:
            A Fixed-Size Encoding Method for Variable-Length Sequences with its
            Application to Neural Network Language Models

        Parameters
        ----------
        peptides : list of strings

        alpha: float
            Forgetting factor

        bidirectional: boolean
            Whether to do both a forward pass and a backward pass over each
            peptide
        """
        # don't try to do length validation since we're allowed to have
        # multiple peptide lengths in a FOFE encoding
        peptides = self.prepare_sequences(peptides)
        n_peptides = len(peptides)
        index_dict = self.index_dict
        n_symbols = len(index_dict)
        if bidirectional:
            result = np.zeros((n_peptides, 2 * n_symbols), dtype=float)
        else:
            result = np.zeros((n_peptides, n_symbols), dtype=float)
        for i, p in enumerate(peptides):
            l = len(p)
            for j, amino_acid in enumerate(p):
                aa_idx = index_dict[amino_acid]
                result[i, aa_idx] += alpha ** (l - j - 1)
                if bidirectional:
                    result[i, n_symbols + aa_idx] += alpha ** j
        return result



from pepnet import SequenceInput, Output, Predictor
from nose.tools import eq_

def test_fixed_length_hotshot():
    model = Predictor(
        inputs=SequenceInput(length=9, variable_length=False, encoding="onehot"),
        outputs=Output(1, activation="sigmoid"))
    seqs = ["A" * 9, "L" * 9]
    y = model.predict(seqs)
    eq_(len(y), 2)

def test_fixed_length_embedding_network():
    model = Predictor(
        inputs=SequenceInput(length=9, variable_length=False, encoding="index"),
        outputs=Output(1, activation="sigmoid"))
    seqs = ["A" * 9, "L" * 9]
    y = model.predict(seqs)
    eq_(len(y), 2)



from pepnet.encoder import Encoder
from nose.tools import eq_
import numpy as np

def test_encoder_index_lists():
    encoder = Encoder()
    S_idx = encoder.index_dict["S"]
    A_idx = encoder.index_dict["A"]
    index_lists = encoder.encode_index_lists(["SSS", "AAA", "SAS"])
    eq_(index_lists, [
        [S_idx, S_idx, S_idx],
        [A_idx, A_idx, A_idx],
        [S_idx, A_idx, S_idx]
    ])

def test_encoder_prepare_sequences_padding():
    encoder = Encoder()
    eq_(encoder.prepare_sequences(["SISI"], 5), ["SISI-"])

def test_encoder_prepare_sequences_start_token():
    encoder = Encoder(add_start_tokens=True)
    eq_(encoder.prepare_sequences(["SISI"], 5), ["^SISI-"])


def test_encoder_prepare_sequences_stop_token():
    encoder = Encoder(add_stop_tokens=True)
    eq_(encoder.prepare_sequences(["SISI"], 5), ["SISI$-"])


def test_encoder_index_array():
    encoder = Encoder()
    S_idx = encoder.index_dict["S"]
    A_idx = encoder.index_dict["A"]
    assert S_idx > 0
    assert A_idx > 0
    X = encoder.encode_index_array(["SSS", "AAA", "SASA"], max_peptide_length=4)
    expected = np.array([
        [S_idx, S_idx, S_idx, 0],
        [A_idx, A_idx, A_idx, 0],
        [S_idx, A_idx, S_idx, A_idx]
    ])
    assert (X == expected).all()


def test_encoder_FOFE():
    # turn off the gap character '-' used for ends of shorter sequences
    encoder = Encoder(variable_length_sequences=False)
    x = encoder.encode_FOFE(["AAA", "SSS", "SASA"])
    eq_(x.shape, (3, 20))

def test_encoder_FOFE_bidirectional():
    # turn off the gap character '-' used for ends of shorter sequences
    encoder = Encoder(variable_length_sequences=False)
    x = encoder.encode_FOFE(["AAA", "SSS", "SASA"], bidirectional=True)
    eq_(x.shape, (3, 40))



from pepnet import Predictor, SequenceInput, Output
import numpy as np

def test_embedding_conv_1_layer():
    model = Predictor(
        inputs=SequenceInput(
            length=3, variable_length=False, conv_filter_sizes={2: 3}),
        outputs=Output(1, activation="sigmoid"))
    X = ["SAY", "FFQ"]
    Y = np.array([True, False])
    model.fit(X, Y)



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from random import shuffle
import numpy as np
from collections import Counter

def _group_sequences_by_kmers(seqs, k=9):
    kmer_to_group_dict = {}
    group_list = []
    for seq in seqs:
        kmers = [seq[i:i + k] for i in range(len(seq) - k + 1)]
        group = None
        for kmer in kmers:
            if kmer in kmer_to_group_dict:
                group = kmer_to_group_dict[kmer]
                break
        if not group:
            group = set([seq])
            group_list.append(group)
        else:
            group.add(seq)
        for kmer in kmers:
            kmer_to_group_dict[kmer] = group
    return group_list

def random_iter(seq):
    shuffled = list(seq)
    shuffle(shuffled)
    return iter(shuffled)

def _group_similar_sequences_and_flatten(seqs, k=9):
    groups = _group_sequences_by_kmers(seqs, k=k)
    seq_list = []
    group_id_list = []
    for i, group in enumerate(random_iter(groups)):
        for seq in random_iter(group):
            seq_list.append(seq)
            group_id_list.append(i)
    return seq_list, np.array(group_id_list)

def group_similar_sequences(seqs, k=9):
    """
    Group sequences by kmer content. Returns list of sequences, group IDs for
    each sequence, and weights that are inversely proportional to size of each
    group.
    """
    seqs, group_ids = _group_similar_sequences_and_flatten(seqs=seqs, k=k)
    weights = np.ones(len(seqs))
    counts = Counter()
    for group_id in group_ids:
        counts[group_id] += 1
    for i, group_id in enumerate(group_ids):
        weights[i] = 1.0 / counts[group_id]
    return seqs, group_ids, weights



from .numeric_input import NumericInput
from .output import Output
from .sequence_input import SequenceInput
from .predictor import Predictor
from .encoder import Encoder

__all__ = [
    "NumericInput",
    "SequenceInput",
    "Output",
    "Predictor",
    "Encoder",
]

__version__ = "0.3.1"



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import keras.backend as K
import theano.tensor as T

def positive_only_mse(y_true, y_pred):
    """
    Mean-squared error loss function that ignores negative values of y_pred.
    Using this as a stop-gap until I can figure out how to avoid the mess
    of explicitly passing an output mask as an Input to a keras model.
    """
    diff = y_pred - y_true
    squared = K.square(diff)
    mask = y_pred >= 0
    squared *= K.cast(mask, "float32")
    return K.mean(squared, axis=-1)

def masked_mse(y_true, y_pred):
    mask = T.isnan(y_true)
    diff = y_pred - y_true
    squared = K.square(diff)
    sum_squared_error = K.sum(
        K.switch(mask, 0.0, squared),
        axis=-1)
    n_valid_per_sample = K.sum(~mask, axis=-1)
    return sum_squared_error / n_valid_per_sample

def masked_binary_crossentropy(y_true, y_pred):
    mask = T.isnan(y_true)
    cross_entropy_values = K.binary_crossentropy(
        output=y_pred,
        target=y_true)
    sum_cross_entropy_values = K.sum(
        K.switch(mask, 0.0, cross_entropy_values), axis=-1)
    n_valid_per_sample = K.sum(~mask, axis=-1)
    return sum_cross_entropy_values / n_valid_per_sample



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .nn_helpers import dense_layers, dense
from .numeric import Numeric
from .losses import masked_mse, masked_binary_crossentropy

class Output(Numeric):
    """
    Numeric output for neural network.
    """
    def __init__(
            self,
            dim=1,
            activation="linear",
            name=None,
            loss="mse",
            dense_layer_sizes=[],
            dense_activation="relu",
            dense_dropout=0,
            dense_batch_normalization=False,
            transform=None,
            inverse_transform=None):
        Numeric.__init__(
            self,
            name=name,
            dim=dim,
            dense_layer_sizes=dense_layer_sizes,
            dense_activation=dense_activation,
            dense_dropout=dense_dropout,
            dense_batch_normalization=dense_batch_normalization,
            transform=transform)
        self.activation = activation
        self.loss = loss
        self.inverse_transform = inverse_transform

    def build(self, value):
        hidden = dense_layers(
            value,
            layer_sizes=self.dense_layer_sizes,
            activation=self.dense_activation,
            dropout=self.dense_dropout,
            batch_normalization=self.dense_batch_normalization)
        output = dense(
            hidden,
            dim=self.dim,
            activation=self.activation,
            name=self.name)
        return output

    def decode(self, x):
        if self.inverse_transform is not None:
            return self.inverse_transform(x)
        return x

    @property
    def loss_fn(self):
        """
        If output requires masking then apply it to the loss function,
        otherwise just return the loss function.
        """
        if self.loss == "mse":
            return masked_mse
        elif self.loss == "binary_crossentropy":
            return masked_binary_crossentropy
        else:
            raise ValueError("Only MSE supported, not '%s'" % self.loss)



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict

from six import string_types
import numpy as np
from keras.models import Model
from keras.utils import plot_model
from serializable import Serializable
import ujson

from .numeric_input import NumericInput
from .sequence_input import SequenceInput
from .output import Output
from .nn_helpers import merge, dense_layers, tensor_shape


class Predictor(Serializable):
    def __init__(
            self,
            inputs,
            outputs,
            merge_mode="concat",
            dense_layer_sizes=[],
            dense_activation="relu",
            dense_dropout=0.25,
            dense_batch_normalization=False,
            optimizer="rmsprop",
            training_metrics=[]):

        if isinstance(inputs, (NumericInput, SequenceInput)):
            inputs = [inputs]
        elif isinstance(inputs, dict):
            inputs_dict = inputs
            inputs = []
            for (name, i) in sorted(inputs_dict.items()):
                if i.name is None:
                    i.name = name
                elif i.name != name:
                    raise ValueError("Input named '%s' given key '%s'" % (i.name, name))
                inputs.append(i)

        if isinstance(outputs, (Output,)):
            outputs = [outputs]
        elif isinstance(outputs, dict):
            outputs_dict = outputs
            outputs = []
            for (name, o) in sorted(outputs_dict.items()):
                if o.name is None:
                    o.name = name
                elif o.name != name:
                    raise ValueError("Output named '%s' given key '%s'" % (o.name, name))
                outputs.append(o)

        if len(outputs) > 1 and any(not o.name for o in outputs):
            raise ValueError("Predictors with multiple outputs must have names for each output")
        if len(outputs) > len({o.name for o in outputs}):
            raise ValueError("All outputs must have distinct names")

        self.inputs = inputs
        self.outputs = outputs
        self.merge_mode = merge_mode
        self.dense_layer_sizes = dense_layer_sizes
        self.dense_activation = dense_activation
        self.dense_dropout = dense_dropout
        self.dense_batch_normalization = dense_batch_normalization
        self.optimizer = optimizer
        self.training_metrics = training_metrics
        self.model = self._build_and_compile()

    @property
    def use_input_dict(self):
        if self.num_inputs == 1:
            input_name = self.input_order[0]
            # if our single input doesn't have a name then don't try to
            # pass a dict to Keras
            return (input_name is not None) and (len(input_name) > 0)
        else:
            if any(
                    (name is None or len(name) == 0)
                    for name in self.input_order):
                raise ValueError("Predictor must have names for all %d inputs" % (
                    self.num_inputs))
            return True

    @property
    def inputs_dict(self):
        return {i.name: i for i in self.inputs}

    @property
    def input_order(self):
        return [i.name for i in self.inputs]

    @property
    def outputs_dict(self):
        return {o.name: o for o in self.outputs}

    @property
    def output_order(self):
        return [o.name for o in self.outputs]

    @property
    def output_names(self):
        return self.output_order

    @property
    def use_output_dict(self):
        if self.num_outputs == 1:
            output_name = self.output_order[0]
            # if our single input doesn't have a name then don't try to
            # pass a dict to Keras
            return (output_name is not None) and (len(output_name) > 0)
        else:
            if any(
                    (name is None or len(name) == 0)
                    for name in self.output_order):
                raise ValueError(
                        "Predictor must have names for all %d outputs" % (
                            self.num_outputs))
            return True

    def _get_single_output(self):
        """
        When use_output_dict is False then we know that there's only one
        output and we can use it without knowing its name.
        """
        outputs = self.outputs
        if len(outputs) == 0:
            raise ValueError("Expected at least one output")
        elif len(outputs) > 1:
            raise ValueError("Expected only one output but got %d" % (
                len(outputs)))
        return outputs[0]

    def _build(self):
        input_dict = {}
        subgraphs_dict = OrderedDict()
        for (input_name, input_descriptor) in self.inputs_dict.items():
            input_obj, subgraph = input_descriptor.build()
            input_dict[input_name] = input_obj
            subgraphs_dict[input_name] = subgraph

        if len(subgraphs_dict) == 0:
            raise ValueError("At least one input required")
        elif isinstance(self.merge_mode, string_types):
            merge_dict = {tuple(subgraphs_dict.keys()): self.merge_mode}
        elif not isinstance(self.merge_mode, dict):
            raise TypeError("Expected 'merge' to be str or dict but got %s : %s " % (
                self.merge_mode, type(self.merge_mode)))
        else:
            merge_dict = self.merge_mode

        combined_tensors = []
        for (input_names, merge_mode) in merge_dict.items():
            if isinstance(input_names, string_types):
                input_names = (input_names,)
            combined_tensors.append(
                merge([subgraphs_dict[name] for name in input_names],
                        merge_mode))
        # concatenate final tensor in case we're combining more than one
        # group
        combined = merge(combined_tensors, "concat")

        hidden = dense_layers(
            combined,
            layer_sizes=self.dense_layer_sizes,
            activation=self.dense_activation,
            dropout=self.dense_dropout,
            batch_normalization=self.dense_batch_normalization)

        output_dict = {}
        for output_name, output_descriptor in self.outputs_dict.items():
            output_graph = output_descriptor.build(hidden)
            output_dict[output_name] = output_graph

        inputs = [input_dict[name] for name in self.input_order]
        outputs = [output_dict[name] for name in self.output_order]
        return Model(inputs=inputs, outputs=outputs)

    def _compile(self, model):
        if self.use_output_dict:
            loss = {
                output.name: output.loss_fn for output in self.outputs
            }
        else:
            loss = self._get_single_output().loss_fn
        model.compile(loss=loss, optimizer=self.optimizer, metrics=self.training_metrics)

    def _build_and_compile(self):
        if self.num_inputs == 0:
            raise ValueError("Predictor must have at least one output")
        model = self._build()
        self._compile(model)
        return model

    @property
    def num_inputs(self):
        return len(self.input_order)

    @property
    def num_outputs(self):
        return len(self.output_order)

    ############################################################################
    #
    # Prediction
    #
    ############################################################################

    def _prepare_inputs(self, inputs):
        """
        Returns dictionary of input name -> input value if use_input_dict is
        True else, returns just encoded representation of single input.
        """
        if isinstance(inputs, (list, np.ndarray)):
            if self.num_inputs != 1:
                raise ValueError("Expected %d inputs but got 1" % self.num_inputs)
            inputs = {self.input_order[0]: inputs}
        elif not isinstance(inputs, dict):
            raise TypeError(
                "Expected inputs to be list, array, or dict, got %s" % (
                    type(inputs)))
        encoded_inputs = {
            name: i.encode(inputs[name])
            for name, i in self.inputs_dict.items()
        }
        lengths = {name: len(x) for (name, x) in encoded_inputs.items()}
        if any(l != list(lengths.values())[0] for l in lengths.values()):
            raise ValueError("All inputs must be of the same length, given %s" % (
                lengths,))
        if self.use_input_dict:
            return encoded_inputs
        else:
            return list(encoded_inputs.values())[0]

    def _prepare_outputs(self, outputs, encode=False, decode=False):
        """
        Returns a dictionary from output name to array of output values.
        """
        if isinstance(outputs, list):
            outputs = np.array(outputs).squeeze().T

        if isinstance(outputs, np.ndarray):
            if outputs.ndim == 1:
                outputs = np.expand_dims(outputs, 1)

            n_given_outputs = outputs.shape[1]

            if self.num_outputs != n_given_outputs:
                raise ValueError("Expected %d outputs but got %d" % (
                    self.num_outputs,
                    n_given_outputs))

            outputs = {
                output_name: outputs[:, i]
                for i, output_name
                in enumerate(self.output_order)
            }
        elif not isinstance(outputs, dict):
            raise ValueError("Expected outputs to list, array, or dict, got %s" % (
                type(outputs)))
        if encode:
            outputs = {
                name: output.encode(outputs[name])
                for name, output in self.outputs_dict.items()
            }
        if decode:
            outputs = {
                name: output.decode(outputs[name])
                for name, output in self.outputs_dict.items()
            }
        lengths = {name: len(x) for (name, x) in outputs.items()}
        if any(l != list(lengths.values())[0] for l in lengths.values()):
            raise ValueError(
                "All outputs must be of the same length, given %s" % (lengths,))
        if self.use_output_dict:
            return outputs
        else:
            return list(outputs.values())[0]

    def predict_scores(self, inputs):
        return self._prepare_outputs(
            self.model.predict(self._prepare_inputs(inputs)),
            decode=False)

    def predict(self, inputs):
        return self._prepare_outputs(
            self.model.predict(self._prepare_inputs(inputs)),
            decode=True)

    ############################################################################
    #
    # Weight estimation
    #
    ############################################################################

    def _prepare_sample_weights(self, sample_weight):
        if sample_weight is not None:
            if self.use_input_dict and len(self.outputs) > 1:
                if isinstance(sample_weight, np.ndarray):
                    sample_weight = {
                        o.name: sample_weight
                        for o in self.outputs
                    }
        return sample_weight

    def _prepare_data_tuple(self, data_tuple):
        """
        Data generators return either (input, output) or
        (input, output, weights) tuples. This function transforms
        these elements for use with Keras.
        """
        if len(data_tuple) == 2:
            inputs, outputs = data_tuple
            weights = None
        else:
            assert len(data_tuple) == 3, \
                "Dataset expected to be (X, Y, weights), got %d elements" % (
                    len(data_tuple),)
            inputs, outputs, weights = data_tuple
        inputs = self._prepare_inputs(inputs)
        outputs = self._prepare_outputs(outputs)
        weights = self._prepare_sample_weights(weights)
        return inputs, outputs, weights

    def _wrap_data_generator(self, generator):
        for data_tuple in generator:
            yield self._prepare_data_tuple(data_tuple)

    def fit(self,
            inputs,
            outputs,
            batch_size=32,
            epochs=100,
            sample_weight=None,
            class_weight=None,
            validation_data=None,
            shuffle=True,
            callbacks=[]):
        inputs = self._prepare_inputs(inputs)
        outputs = self._prepare_outputs(outputs, encode=True)
        sample_weight = self._prepare_sample_weights(sample_weight)

        if validation_data is not None:
            validation_data = self._prepare_data_tuple(validation_data)

        return self.model.fit(
            inputs,
            outputs,
            batch_size=batch_size,
            epochs=epochs,
            sample_weight=sample_weight,
            class_weight=class_weight,
            shuffle=shuffle,
            validation_data=validation_data,
            callbacks=callbacks)

    def fit_generator(
            self,
            generator,
            steps_per_epoch,
            **kwargs):
        """
        Expects a generator which returns tuples of
            (inputs, outputs)
            -or-
            (inputs, outputs, sample_weights)
        which are then transformed appropriately before being
        passed on to the fit_generator method of the underlying
        Keras model.
        """
        return self.model.fit_generator(
            generator=self._wrap_data_generator(generator),
            steps_per_epoch=steps_per_epoch,
            **kwargs)

    ############################################################################
    #
    # Model visualization
    #
    ############################################################################

    def save_diagram(self, filename="model.png"):
        plot_model(self.model, to_file=filename)

    ############################################################################
    #
    # Serialization methods and related helpers
    #
    ############################################################################

    def _input_to_repr(self, input_obj):
        """
        Return a serializable representation of an input.
        """
        return (input_obj.__class__.__name__, input_obj.to_dict())

    @classmethod
    def _input_from_repr(cls, input_repr):
        """
        Create an input from a flattened representation
        """
        name, config = input_repr
        if name == "SequenceInput":
            return SequenceInput.from_dict(config)
        elif name == "NumericInput":
            return NumericInput.from_dict(config)
        else:
            raise ValueError("Invalid input class: %s" % (name,))

    def _output_to_repr(self, output_obj):
        return (output_obj.__class__.__name__, output_obj.to_dict())

    @classmethod
    def _output_from_repr(self, output_repr):
        name, config = output_repr
        if name == "Output":
            return Output.from_dict(config)
        else:
            raise ValueError("Invalid output name: %s" % (name,))

    def get_weights(self):
        return [w.get_value().squeeze() for w in self.model.weights]

    def set_weights(self, weights):
        if len(self.model.weights) != len(weights):
            raise ValueError("Expected %d weight arrays but got %d" % (
                len(self.model.weights),
                len(weights)))
        for w_tensor, w_values in zip(self.model.weights, weights):
            shape = tensor_shape(w_tensor)
            w_compatible = w_values.reshape(shape)
            w_compatible = w_compatible.astype(w_tensor.dtype)
            w_tensor.set_value(w_compatible)

    def to_dict(self):
        return {
            "inputs": [self._input_to_repr(i) for i in self.inputs],
            "outputs": [self._output_to_repr(o) for o in self.outputs],
            "merge_mode": self.merge_mode,
            "dense_layer_sizes": self.dense_layer_sizes,
            "dense_activation": self.dense_activation,
            "dense_dropout": self.dense_dropout,
            "dense_batch_normalization": self.dense_batch_normalization,
            "optimizer": self.optimizer,
            "training_metrics": self.training_metrics,
            "model_weights": [w.tolist() for w in self.get_weights()]
        }

    @classmethod
    def from_dict(self, config_dict):
        model_weights_as_lists = config_dict.pop("model_weights")
        model_weights = [np.array(values) for values in model_weights_as_lists]
        input_reprs = config_dict.pop("inputs")
        output_reprs = config_dict.pop("outputs")
        inputs = [self._input_from_repr(i) for i in input_reprs]
        outputs = [self._output_from_repr(o) for o in output_reprs]
        predictor = Predictor(inputs=inputs, outputs=outputs, **config_dict)
        predictor.set_weights(model_weights)
        return predictor

    def to_json(self):
        return ujson.dumps(self.to_dict())

    def to_json_file(self, filename):
        with open(filename, "w") as f:
            f.write(self.to_json())

    @classmethod
    def from_json(cls, json_string):
        config_dict = ujson.loads(json_string)
        return cls.from_dict(config_dict)

    @classmethod
    def from_json_file(cls, filename):
        with open(filename, "r") as f:
            s = f.read()
            if len(s) == 0:
                raise ValueError("File '%s' is empty" % filename)
            return cls.from_json(s)



from pepnet import Predictor, SequenceInput, Output

import numpy as np

def test_model_with_fixed_length_context():
    model = Predictor(
        inputs={
            "upstream": SequenceInput(length=1, variable_length=False),
            "downstream": SequenceInput(length=1, variable_length=False),
            "peptide": SequenceInput(length=3, variable_length=True)},
        outputs=Output(1, activation="sigmoid"))

    Y = np.array([True, False, True, False])
    input_dict = {
        "upstream": ["Q", "A", "L", "I"],
        "downstream": ["S"] * 4,
        "peptide": ["SYF", "QQ", "C", "GLL"]
    }
    model.fit(input_dict, Y, epochs=20)
    Y_pred = model.predict(input_dict)
    assert (Y == (Y_pred > 0.5)).all(), (Y, Y_pred)



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import (
    print_function,
    division,
    absolute_import,
)

###
# 20 common amino acids
###
amino_acids_dict = {
    "A": "Alanine",
    "R": "Arginine",
    "N": "Asparagine",
    "D": "Aspartic Acid",
    "C": "Cysteine",
    "E": "Glutamic Acid",
    "Q": "Glutamine",
    "G": "Glycine",
    "H": "Histidine",
    "I": "Isoleucine",
    "L": "Leucine",
    "K": "Lysine",
    "M": "Methionine",
    "F": "Phenylalanine",
    "P": "Proline",
    "S": "Serine",
    "T": "Threonine",
    "W": "Tryptophan",
    "Y": "Tyrosine",
    "V": "Valine",
}

###
# Including 7 post-translational modifications commonly found by mass spec
###
post_translation_modifications_dict = {
    "s": "Phospho-Serine",
    "t": "Phospho-Threonine",
    "y": "Phospho-Tyrosine",
    "c": "Cystine",
    "m": "Methionine sulfoxide",
    "q": "Pyroglutamate",
    "n": "Pyroglutamic acid"
}

def amino_acids_with_ptms(ptms=[]):
    """
    Generate a dictionary mapping amino acids to their names, along
    with a specific set of post-translational modifications.
    """
    amino_acids_with_ptms_dict = amino_acids_dict.copy()
    for ptm in ptms:
        name = post_translation_modifications_dict[ptm]
        amino_acids_with_ptms_dict[ptm] = name
    return amino_acids_with_ptms_dict



# Copyright (c) 2017. Mount Sinai School of Medicine
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from serializable import Serializable

class Numeric(Serializable):
    """
    Base class for numeric input and output
    """
    def __init__(
            self,
            dim,
            name=None,
            dtype="float32",
            dense_layer_sizes=[],
            dense_activation="relu",
            dense_dropout=0,
            dense_batch_normalization=False,
            transform=None):
        """
        Parameters
        ----------
        dim : int
            Number of input dimensions

        name : str
            Name of input sequence

        dtype : str
            Most common option is "float32" but might also be "int32"

        dense_layer_sizes : list of int
            Size of each dense layer after the input

        dense_activation : str
            Activation functin for dense layers after input

        dense_dropout : float
            Fraction of values to randomly set to 0 during training

        dense_batch_normalization : bool
            Use Batch Normalization after hidden layers

        transform : fn, optional
            Function to transform elements of numeric input/output
        """
        self.name = name
        self.dim = dim
        self.dtype = dtype
        self.dense_layer_sizes = dense_layer_sizes
        self.dense_activation = dense_activation
        self.dense_dropout = dense_dropout
        self.dense_batch_normalization = dense_batch_normalization
        self.transform = transform

    def build(self):
        raise NotImplementedError("Numeric cannot be directly instantiated")

    def encode(self, x):
        if self.transform:
            return self.transform(x)
        return x



import pandas
from numpy.random import randn
from numpy import log, exp

from pepnet import Predictor, SequenceInput, NumericInput, Output
from pepnet.synthetic_data import synthetic_peptides_by_subsequence
from nose.tools import eq_


def test_simple_numeric_predictor():
    predictor = Predictor(
        inputs=[NumericInput(dim=30)],
        outputs=[Output(dim=1, activation="sigmoid")],
        dense_layer_sizes=[30],
        dense_activation="relu")
    y = predictor.predict(randn(10, 30))
    eq_(len(y), 10)

def test_simple_numeric_predictor_named():
    predictor = Predictor(
        inputs=[NumericInput(name="x", dim=30)],
        outputs=[Output(dim=1, name="y", activation="sigmoid")],
        dense_layer_sizes=[30],
        dense_activation="relu")
    y = predictor.predict({"x": randn(10, 30)})["y"]
    eq_(len(y), 10)


def test_simple_sequence_predictor():
    predictor = Predictor(
        inputs=[SequenceInput(length=4, variable_length=True)],
        outputs=[Output(dim=1, activation="sigmoid")],
        dense_layer_sizes=[30],
        dense_activation="relu")
    y = predictor.predict(["SFY-"] * 10)
    eq_(len(y), 10)

def test_simple_sequence_predictor_named():
    predictor = Predictor(
        inputs=[SequenceInput(length=4, name="x", variable_length=True)],
        outputs=[Output(dim=1, activation="sigmoid", name="y")],
        dense_layer_sizes=[30],
        dense_activation="relu")
    y = predictor.predict({"x": ["SFY-"] * 10})["y"]
    eq_(len(y), 10)

def test_two_input_predictor():
    predictor = Predictor(
        inputs=[
            SequenceInput(length=4, name="x1", variable_length=True),
            NumericInput(dim=30, name="x2")],
        outputs=[Output(name="y", dim=1, activation="sigmoid")],
        dense_layer_sizes=[30],
        dense_activation="relu")
    y = predictor.predict({
        "x1": ["SFY-"] * 10,
        "x2": randn(10, 30)})["y"]
    eq_(len(y), 10)

def test_predictor_output_transform():
    predictor = Predictor(
        inputs=[NumericInput(dim=30, name="x")],
        outputs=[
            Output(
                name="y",
                dim=1,
                activation="sigmoid",
                transform=log,
                inverse_transform=exp)],
        dense_layer_sizes=[30],
        dense_activation="relu")
    y = predictor.predict({"x": randn(10, 30)})["y"]
    eq_(len(y), 10)
    # make sure transformed outputs are within given bounds
    assert exp(0.0) <= y.min() <= exp(1.0)
    assert exp(0.0) <= y.max() <= exp(1.0)

def test_predictor_on_more_data():
    predictor = Predictor(
        inputs=[SequenceInput(length=20, name="x", variable_length=True)],
        outputs=[Output(dim=1, activation="sigmoid", name="y")],
        dense_layer_sizes=[30],
        dense_activation="relu")

    train_df = synthetic_peptides_by_subsequence(1000)
    test_df = synthetic_peptides_by_subsequence(1000)
    predictor.fit(
        {"x": train_df.index.values}, train_df.binder.values, epochs=20)
    y_pred = predictor.predict({"x": test_df.index.values})['y']
    y_pred = pandas.Series(y_pred, index=test_df.index)
    binder_mean_pred = y_pred[test_df.binder].mean()
    nonbinder_mean_pred = y_pred[~test_df.binder].mean()
    print(binder_mean_pred, nonbinder_mean_pred)
    assert binder_mean_pred > nonbinder_mean_pred * 2, (
        binder_mean_pred, nonbinder_mean_pred)



from pepnet.amino_acids import amino_acids_with_ptms
from nose.tools import eq_

def test_ptm_dict_no_args():
    alphabet = amino_acids_with_ptms([])
    eq_(len(alphabet), 20)

def test_ptm_dict_two_args():
    alphabet = amino_acids_with_ptms(["c", "m"])
    eq_(len(alphabet), 22)



from pepnet.nn_helpers import highway_layers
from keras.layers import Input, Embedding, Dense, Flatten
from keras.models import Model
import numpy as np

def test_highway_layers():
    n_highway_layers = 5
    x = Input(shape=(8,), dtype="int32")
    v = Embedding(input_dim=2, output_dim=10)(x)
    v = Flatten()(v)
    assert hasattr(v, "_keras_shape")
    v = highway_layers(v, n_layers=n_highway_layers)
    output = Dense(1)(v)
    model = Model(inputs=[x], outputs=[output])
    assert len(model.layers) > n_highway_layers * 3
    x = np.array([
        [1] + [0] * 7,
        [0] * 8,
        [0] * 7 + [1]])
    y = np.array([0, 1, 0])
    model.compile("rmsprop", "mse")
    model.fit(x, y, epochs=10)
    pred = model.predict(x)
    mean_diff = np.abs(pred - y).mean()
    assert mean_diff < 0.5, pred



from pepnet import Predictor, SequenceInput, Output
from nose.tools import eq_

def test_basic_rnn():
    pred = Predictor(
        inputs=SequenceInput(
            name="x",
            length=4,
            variable_length=True,
            encoding="index",
            rnn_layer_sizes=[20],
            rnn_type="lstm",
            rnn_bidirectional=True),
        outputs=Output(dim=1, activation="sigmoid", name="y"))
    x = ["SF", "Y", "AALL"]
    y = pred.predict({"x": x})["y"]
    eq_(len(x), len(y))
    found_rnn_layer = any(
        "bidirectional" in layer.name for layer in pred.model.layers)
    assert found_rnn_layer



import numpy
import pandas

from ..amino_acids import amino_acids_dict
from .random_peptides import random_peptides

AMINO_ACIDS = list(amino_acids_dict)


def synthetic_peptides_by_subsequence(
        num_peptides,
        fraction_binders=0.5,
        lengths=range(8, 20),
        binding_subsequences=["A?????Q"]):
    """
    Generate a toy dataset where each peptide is a binder if and only if it
    has one of the specified subsequences.
    
    Parameters
    ----------
    num_peptides : int
        Number of rows in result
    
    fraction_binders : float
        Fraction of rows in result where "binder" col is 1
    
    lengths : dict, Series, or list
        If a dict or Series, then this should map lengths to the fraction of the
        result to have the given peptide length. If it's a list of lengths then
        all lengths are given equal weight.
    
    binding_subsequences : list of string
        Peptides with any of the given subsequences will be considered binders.
        Question marks ("?") in these sequences will be replaced by random
        amino acids.
        
    Returns
    ----------
    pandas.DataFrame, indexed by peptide sequence. The "binder" column is a
    binary indicator for whether the peptide is a binder.
    """
    if not isinstance(lengths, dict):
        lengths = dict((length, 1.0) for length in lengths)

    lengths_series = pandas.Series(lengths)
    lengths_series /= len(lengths)

    num_binders = int(round(num_peptides * fraction_binders))
    num_non_binders = num_peptides - num_binders
    print(num_binders, num_non_binders)

    peptides = []

    # Generate non-binders
    for (length, weight) in lengths_series.iteritems():
        peptides.extend(
            random_peptides(round(weight * num_non_binders), round(length)))

    for binding_core in binding_subsequences:
        # Generate binders
        lengths_binders = lengths_series.ix[
            lengths_series.index >= len(binding_core)
        ]
        normalized_lengths_binders = (
            lengths_binders /
            lengths_binders.sum() /
            len(binding_subsequences))

        for (length, weight) in normalized_lengths_binders.iteritems():
            if length >= len(binding_core):
                num_peptides_to_make = int(round(weight * num_binders))
                if length == len(binding_core):
                    start_positions = [0] * num_peptides_to_make
                else:
                    start_positions = numpy.random.choice(
                        length - len(binding_core), num_peptides_to_make)
                peptides.extend(
                    "".join([
                        random_peptides(1, length=start_position)[0],
                        binding_core,
                        random_peptides(1, length=length - len(
                            binding_core) - start_position)[0],
                    ])
                    for start_position in start_positions)

    df = pandas.DataFrame(index=set(peptides))
    df["binder"] = False
    for binding_core in binding_subsequences:
        df["binder"] = df["binder"] | df.index.str.contains(
            binding_core,
            regex=False)

    def replace_question_marks(s):
        while "?" in s:
            s = s.replace("?", numpy.random.choice(AMINO_ACIDS))
        return s

    df.index = df.index.map(replace_question_marks)
    df_shuffled = df.sample(frac=1)
    return df_shuffled
